﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.EvolutionTracking;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic {
  [Item("TraceCalculator", "Walks a genealogy graph and produces a trace of the specified subtree")]
  [StorableClass]
  public class TraceCalculator : Item {
    private readonly IGenealogyGraph<ISymbolicExpressionTree> traceGraph;
    private readonly Dictionary<IGenealogyGraphNode<ISymbolicExpressionTree>, Tuple<int, int>> traceMap;

    public IGenealogyGraph<ISymbolicExpressionTree> TraceGraph { get { return traceGraph; } }

    public TraceCalculator() {
      traceGraph = new GenealogyGraph<ISymbolicExpressionTree>();
      traceMap = new Dictionary<IGenealogyGraphNode<ISymbolicExpressionTree>, Tuple<int, int>>();
    }

    protected TraceCalculator(TraceCalculator original, Cloner cloner)
      : base(original, cloner) {
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new TraceCalculator(this, cloner);
    }

    public static IGenealogyGraph<ISymbolicExpressionTree> TraceSubtree(IGenealogyGraphNode<ISymbolicExpressionTree> node, int subtreeIndex) {
      var tc = new TraceCalculator();
      tc.Trace(node, subtreeIndex);
      return tc.TraceGraph;
    }

    public IGenealogyGraph<ISymbolicExpressionTree> Trace(IGenealogyGraphNode<ISymbolicExpressionTree> node, int subtreeIndex) {
      traceGraph.Clear();
      traceMap.Clear();
      Trace(node, subtreeIndex, null);
      return traceGraph;
    }

    /// <summary>
    /// This method starts from a given vertex in the genealogy graph and works its way 
    /// up the ancestry trying to track the structure of the subtree given by subtreeIndex.
    /// This method will skip genealogy graph nodes that did not have an influence on the 
    /// structure of the tracked subtree. 
    /// 
    /// Only genealogy nodes which did have an influence are added (as copies) to the trace 
    /// and are consequently called 'trace nodes'. 
    /// 
    /// The arcs connecting trace nodes hold information about the locations of the subtrees
    /// and fragments that have been swapped in the form of a tuple (si, fi, lastSi, lastFi),
    /// where:
    /// - si is the subtree index in the current trace node
    /// - fi is the fragment index in the current trace node
    /// - lastSi is the subtree index in the previous trace node
    /// - lastFi is the subtree index in the previous trace node
    /// </summary>
    /// <param name="g">The current node in the genealogy graph</param>
    /// <param name="si">The index of the traced subtree</param>
    /// <param name="last">The last added node in the trace graph</param>
    private void Trace(IGenealogyGraphNode<ISymbolicExpressionTree> g, int si, IGenealogyGraphNode<ISymbolicExpressionTree> last = null) {
      while (g.Parents.Any()) {
        var parents = g.Parents.ToList();
        var fragment = (IFragment<ISymbolicExpressionTreeNode>)g.InArcs.Last().Data;
        if (fragment == null) {
          // the node is either an elite node or (in rare cases) no fragment was transferred
          g = parents[0];
          continue;
        }

        int fi = fragment.Index1;               // fragment index
        int fl = fragment.Root.GetLength();     // fragment length
        int sl = g.Data.NodeAt(si).GetLength(); // subtree length

        #region trace crossover
        if (parents.Count == 2) {
          if (fi == si) {
            g = parents[1];
            si = fragment.Index2;
            continue;
          }
          if (fi < si) {
            if (fi + fl > si) {
              // fragment contains subtree
              g = parents[1];
              si += fragment.Index2 - fi;
            } else {
              // fragment distinct from subtree
              g = parents[0];
              si += g.Data.NodeAt(fi).GetLength() - fl;
            }
            continue;
          }
          if (fi > si) {
            if (fi < si + sl) {
              // subtree contains fragment => branching point in the fragment graph
              var n = GetTraceNode(g, si, fi);
              Trace(parents[0], si, n);
              Trace(parents[1], fragment.Index2, n);
              break;
            } else {
              // subtree and fragment are distinct. 
              g = parents[0];
              continue;
            }
          }
        }
        #endregion
        #region trace mutation
        // mutation is handled in a simple way: we branch every time there is an overlap between the subtree and the fragment
        // (since mutation effects can be quite unpredictable: replace branch, change node, shake tree, etc)
        if (parents.Count == 1) {
          Debug.Assert(fragment.Index1 == fragment.Index2);
          // check if the subtree and the fragment overlap => branch out
          if ((si == fi) || (si < fi && fi < si + sl) || (fi < si && si < fi + fl)) {
            var n = GetTraceNode(g, si, fi);
            int i = si < fi ? si : fi;
            Trace(parents[0], i, n);
            break;
          } else {
            // if they don't overlap, go up
            g = parents[0];
            if (fi < si)
              si += g.Data.NodeAt(fi).GetLength() - fl;
            continue;
          }
        }
        #endregion
        throw new InvalidOperationException("A node cannot have more than two parents");
      }
      // when we are out of the while the last vertex must be connected with the current one
      ConnectLast(g, si, last);
    }

    /// <summary>
    /// Get the trace node from the trace graph which corresponds to node g from the genealogy graph.
    /// If the trace graph does not contain such a node, one is created by performing a shallow copy of g, then inserted into the trace graph.
    /// </summary>
    /// <param name="g">The genealogy graph node</param>
    /// <param name="si">The subtree index</param>
    /// <param name="fi">The fragment index</param>
    /// <returns></returns>
    private IGenealogyGraphNode<ISymbolicExpressionTree> GetTraceNode(IGenealogyGraphNode<ISymbolicExpressionTree> g, int si, int fi) {
      var n = traceGraph.GetByContent(g.Data);
      if (n == null) {
        n = g.Copy();
        traceGraph.AddVertex(n);
        Debug.Assert(!traceMap.ContainsKey(n));
        traceMap[n] = new Tuple<int, int>(si, fi);
      }
      return n;
    }

    /// <summary>
    /// Connect the current node of the trace graph with the node that was previously added (@last). The current node of the trace graph is determined by the content
    /// of the genealogy graph node @g.  
    /// </summary>
    /// <param name="g">The current node in the genealogy graph</param>
    /// <param name="si">The index of the traced subtree</param>
    /// <param name="last">The last added node in the trace graph</param>
    private void ConnectLast(IGenealogyGraphNode<ISymbolicExpressionTree> g, int si, IGenealogyGraphNode<ISymbolicExpressionTree> last) {
      IFragment<ISymbolicExpressionTreeNode> fragment = g.Parents.Any() ? (IFragment<ISymbolicExpressionTreeNode>)g.InArcs.Last().Data : null;
      int fi = fragment == null ? 0 : fragment.Index1; // fragment index
      var n = GetTraceNode(g, si, fi); // will append n to the trace graph if it does not exist
      if (last == null)
        return;
      var lastTraceData = traceMap[last];
      int lastSi = lastTraceData.Item1; // last subtree index
      int lastFi = lastTraceData.Item2; // last fragment index
      var td = new Tuple<int, int, int, int>(si, fi, lastSi, lastFi); // trace data
      var arc = n.InArcs.SingleOrDefault(a => a.Source == last && a.Data.Equals(td));
      if (arc == null) {
        arc = new GenealogyGraphArc(last, n) { Data = td };
        traceGraph.AddArc(arc);
      }
    }
  }

  internal static class Util {
    // shallow node copy (does not clone the data or the arcs)
    public static IGenealogyGraphNode<ISymbolicExpressionTree> Copy(this IGenealogyGraphNode<ISymbolicExpressionTree> node) {
      return new GenealogyGraphNode<ISymbolicExpressionTree>(node.Data) { Rank = node.Rank, Quality = node.Quality };
    }
    #region some helper methods for shortening the tracing code
    public static ISymbolicExpressionTreeNode NodeAt(this ISymbolicExpressionTree tree, int index) {
      return NodeAt(tree.Root, index);
    }
    public static ISymbolicExpressionTreeNode NodeAt(this ISymbolicExpressionTreeNode root, int index) {
      return root.IterateNodesPrefix().ElementAt(index);
    }
    #endregion
  }
}
